#include "AtomicFileUtils.hpp"

#if PLATFORM_WINDOWS
#include <windows.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fileapi.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/file.h>
#endif

using std::string;
using std::vector;

void atomicReadFile(const string& fileName, vector<uint8_t>* pData)
{
    atomicReplaceFile(fileName, pData, *pData);
}

void atomicReadFile(const string& fileName, vector<char>* pData)
{
    atomicReplaceFile(fileName, pData, *pData);
}

void atomicReadFile(const string& fileName, string* pData)
{
    atomicReplaceFile(fileName, pData, *pData);
}

namespace
{
    template<class Data>
    bool atomicReplaceFileImpl(const string& fileName, Data* pOldData, const Data& newData)
    {
#if PLATFORM_WINDOWS
        int fd = _open(fileName.c_str(), _O_RDWR | _O_CREAT | _O_BINARY, _S_IREAD | _S_IWRITE);
        if (fd == -1)
        {
            fprintf(stderr, "atomicReplaceFile: _open failed\n");
            exit(1);
        }

        HANDLE h = (HANDLE)_get_osfhandle(fd);
        if (h == INVALID_HANDLE_VALUE)
        {
            fprintf(stderr, "atomicReplaceFile: _get_osfhandle failed\n");
            exit(1);
        }

        OVERLAPPED overlapped;
        memset(&overlapped, 0, sizeof(overlapped));
        //overlapped.DUMMYUNIONNAME.DUMMYSTRUCTNAME.Offset = 0;
        //overlapped.DUMMYUNIONNAME.DUMMYSTRUCTNAME.OffsetHigh = 0;
        overlapped.Offset = 0;
        overlapped.OffsetHigh = 0;
        overlapped.hEvent = 0;
        if (!LockFileEx(h, LOCKFILE_EXCLUSIVE_LOCK, 0, 0xffffffff, 0xffffffff, &overlapped))
        {
            fprintf(stderr, "atomicReplaceFile: LockFileEx failed\n");
            exit(1);
        }

        size_t curSize;
        {
            LARGE_INTEGER liCurSize;
            if (!GetFileSizeEx(h, &liCurSize))
            {
                fprintf(stderr, "atomicReplaceFile: GetFileSizeEx failed\n");
                exit(1);
            }

            curSize = (size_t)liCurSize.QuadPart;
        }

        Data curData;

        if (curSize != 0)
        {
            curData.resize(curSize);

            size_t pos = 0;
            while (pos < curSize)
            {
                int bytesToRead = curSize - pos > INT_MAX ? INT_MAX : curSize - pos;
                int bytesRead = _read(fd, &curData[pos], bytesToRead);
                if (bytesRead == -1)
                {
                    fprintf(stderr, "atomicReplaceFile: _read failed\n");
                    exit(1);
                }

                if (bytesRead <= 0 || bytesRead > bytesToRead)
                {
                    fprintf(stderr, "atomicReplaceFile: _read size error\n");
                    exit(1);
                }

                pos += (size_t)bytesRead;
            }
        }

        size_t oldSize = pOldData->size();
        size_t newSize = newData.size();

        bool replaced =
            (curSize == oldSize) &&
            (curSize == 0 ||
             memcmp(&curData[0], &(*pOldData)[0], curSize) == 0);

        if (!replaced)
        {
            pOldData->swap(curData);
        }
        else if (oldSize != newSize ||
                 (oldSize != 0 &&
                  memcmp(&(*pOldData)[0], &newData[0], oldSize) != 0))
        {
            {
                long lsize = _lseek(fd, 0, SEEK_SET);
                if (lsize == -1)
                {
                    fprintf(stderr, "atomicReplaceFile: _lseek(SEEK_SET) failed\n");
                    exit(1);
                }

                if (lsize != 0)
                {
                    fprintf(stderr, "atomicReplaceFile: _lseek(SEEK_SET) size error\n");
                    exit(1);
                }
            }

            if (newSize != 0)
            {
                size_t pos = 0;
                while (pos < newSize)
                {
                    unsigned int bytesToWrite = newSize - pos > INT_MAX ? INT_MAX : newSize - pos;
                    int bytesWritten = _write(fd, &newData[pos], bytesToWrite);

                    if (bytesWritten == -1)
                    {
                        fprintf(stderr, "atomicReplaceFile: _write failed\n");
                        exit(1);
                    }

                    if (bytesWritten <= 0 || (unsigned int)bytesWritten > bytesToWrite)
                    {
                        fprintf(stderr, "atomicReplaceFile: _write size error\n");
                        exit(1);
                    }

                    pos += (size_t)bytesWritten;
                }
            }

            if (!SetEndOfFile(h))
            {
                fprintf(stderr, "atomicReplaceFile: SetEndOfFile failed\n");
                exit(1);
            }
        }

        if (!UnlockFileEx(h, 0, 0xffffffff, 0xffffffff, &overlapped))
        {
            fprintf(stderr, "atomicReplaceFile: UnlockFileEx failed\n");
            exit(1);
        }

        if (_close(fd) != 0)
        {
            fprintf(stderr, "atomicReplaceFile: _close failed\n");
            exit(1);
        }

        return replaced;
#else
        int fd = open(fileName.c_str(), O_RDWR | O_CREAT | O_CLOEXEC, 0666);
        if (fd == -1)
        {
            fprintf(stderr, "atomicReplaceFile: open failed\n");
            exit(1);
        }

        if (flock(fd, LOCK_EX) != 0)
        {
            fprintf(stderr, "atomicReplaceFile: flock(LOCK_EX) failed\n");
            exit(1);
        }

        size_t curSize;
        {
            off_t osize = lseek(fd, 0, SEEK_END);
            if (osize == (off_t)-1)
            {
                fprintf(stderr, "atomicReplaceFile: lseek(SEEK_END) failed\n");
                exit(1);
            }

            if (osize < 0)
            {
                fprintf(stderr, "atomicReplaceFile: lseek(SEEK_END) size error\n");
                exit(1);
            }

            curSize = (size_t)osize;
        }

        {
            off_t osize = lseek(fd, 0, SEEK_SET);
            if (osize == (off_t)-1)
            {
                fprintf(stderr, "atomicReplaceFile: lseek(SEEK_SET) failed\n");
                exit(1);
            }

            if (osize != 0)
            {
                fprintf(stderr, "atomicReplaceFile: lseek(SEEK_SET) size error\n");
                exit(1);
            }
        }

        Data curData;

        if (curSize != 0)
        {
            curData.resize(curSize);

            size_t pos = 0;
            while (pos < curSize)
            {
                ssize_t bytesRead = read(fd, &curData[pos], curSize - pos);
                if (bytesRead == -1)
                {
                    fprintf(stderr, "atomicReplaceFile: read failed\n");
                    exit(1);
                }

                if (bytesRead <= 0 || (size_t)bytesRead > curSize - pos)
                {
                    fprintf(stderr, "atomicReplaceFile: read size error\n");
                    exit(1);
                }

                pos += (size_t)bytesRead;
            }
        }

        size_t oldSize = pOldData->size();
        size_t newSize = newData.size();

        bool replaced =
            (curSize == oldSize) &&
            (curSize == 0 ||
             memcmp(&curData[0], &(*pOldData)[0], curSize) == 0);

        if (!replaced)
        {
            pOldData->swap(curData);
        }
        else if (oldSize != newSize ||
                 (oldSize != 0 &&
                  memcmp(&(*pOldData)[0], &newData[0], oldSize) != 0))
        {
            {
                off_t osize = lseek(fd, 0, SEEK_SET);
                if (osize == (off_t)-1)
                {
                    fprintf(stderr, "atomicReplaceFile: lseek(SEEK_SET) failed\n");
                    exit(1);
                }

                if (osize != 0)
                {
                    fprintf(stderr, "atomicReplaceFile: lseek(SEEK_SET) size error\n");
                    exit(1);
                }
            }

            if (newSize != 0)
            {
                size_t pos = 0;
                while (pos < newSize)
                {
                    ssize_t bytesWritten = write(fd, &newData[pos], newSize - pos);

                    if (bytesWritten == -1)
                    {
                        fprintf(stderr, "atomicReplaceFile: write failed\n");
                        exit(1);
                    }

                    if (bytesWritten <= 0 || (size_t)bytesWritten > newSize - pos)
                    {
                        fprintf(stderr, "atomicReplaceFile: write size error\n");
                        exit(1);
                    }

                    pos += (size_t)bytesWritten;
                }
            }

            if (ftruncate(fd, newSize) != 0)
            {
                fprintf(stderr, "atomicReplaceFile: ftruncate failed\n");
                exit(1);
            }
        }

        if (flock(fd, LOCK_UN) != 0)
        {
            fprintf(stderr, "atomicReplaceFile: flock(LOCK_UN) failed\n");
            exit(1);
        }

        if (close(fd) != 0)
        {
            fprintf(stderr, "atomicReplaceFile: close failed\n");
            exit(1);
        }

        return replaced;
#endif
    }
}

bool atomicReplaceFile(const string& fileName, vector<uint8_t>* pOldData, const vector<uint8_t>& newData)
{
    return atomicReplaceFileImpl(fileName, pOldData, newData);
}

bool atomicReplaceFile(const string& fileName, vector<char>* pOldData, const vector<char>& newData)
{
    return atomicReplaceFileImpl(fileName, pOldData, newData);
}

bool atomicReplaceFile(const string& fileName, string* pOldData, const string& newData)
{
    return atomicReplaceFileImpl(fileName, pOldData, newData);
}
