#include "Eval.hpp"
#include "StringUtils.hpp"
#include <unordered_map>

using std::string;
using std::vector;
using std::unordered_map;

namespace
{
    uint64_t calcCost(const Problem& problem,
                      uint64_t baseCost,
                      const Rect& rect)
    {
        uint64_t numerator = baseCost * problem.width * problem.height;
        uint64_t denominator = rect.size();

        uint64_t cost = (numerator + denominator/2) / denominator;

        return cost;
    }
}

bool evalMove(const Problem& problem,
              const Move& move,
              Canvas& canvas,
              Image* pResult,
              uint64_t* pCost,
              Grid<uint64_t>* pMoveIdxs,
              uint64_t moveIdx,
              vector<Rect>* pRects,
              string* pMsg)
{
    uint64_t cost = 0;

    switch (move.moveKind)
    {
    case MoveKind::Color:
    {
        uint64_t blockNum = move.blockNums[0];
        if (blockNum >= canvas.blocks.size() ||
            !canvas.blocks[blockNum].valid)
        {
            if (pMsg) *pMsg = strprintf("Move %" PRId64 ": Invalid block number\n", moveIdx);
            return false;
        }

        Rect rect = canvas.blocks[blockNum].rect;

        if (pResult)
        {
            for (uint64_t y = rect.lowY; y < rect.highY; y++)
            {
                for (uint64_t x = rect.lowX; x < rect.highX; x++)
                {
                    (*pResult)(x, y) = move.color;
                }
            }
        }

        if (pMoveIdxs)
        {
            for (uint64_t y = rect.lowY; y < rect.highY; y++)
            {
                for (uint64_t x = rect.lowX; x < rect.highX; x++)
                {
                    (*pMoveIdxs)(x, y) = moveIdx;
                }
            }
        }

        if (pRects)
        {
            if (!pRects->empty() ||
                rect.lowX != 0 ||
                rect.lowY != 0 ||
                rect.highX != problem.width ||
                rect.highY != problem.height)
            {
                pRects->push_back(rect);
            }
        }

        cost = calcCost(problem, 5, rect);

        break;
    }
    case MoveKind::CutX:
    case MoveKind::CutY:
    case MoveKind::CutXY:
    {
        uint64_t blockNum = move.blockNums[0];
        if (blockNum >= canvas.blocks.size() ||
            !canvas.blocks[blockNum].valid)
        {
            if (pMsg) *pMsg = strprintf("Move %" PRId64 ": Invalid block number\n", moveIdx);
            return false;
        }

        Rect rect = canvas.blocks[blockNum].rect;

        if (move.moveKind != MoveKind::CutY)
        {
            if (move.x <= rect.lowX || move.x >= rect.highX)
            {
                if (pMsg) *pMsg = strprintf("Move %" PRId64 ": Invalid X coordinate: %" PRIu64 "\n", moveIdx, move.x);
                return false;
            }
        }

        if (move.moveKind != MoveKind::CutX)
        {
            if (move.y <= rect.lowY || move.y >= rect.highY)
            {
                if (pMsg) *pMsg = strprintf("Move %" PRId64 ": Invalid Y coordinate: %" PRIu64 "\n", moveIdx, move.y);
                return false;
            }
        }

        canvas.blocks[blockNum].valid = false;

        switch (move.moveKind)
        {
        case MoveKind::CutX:
        {
            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = Rect(rect.lowX, rect.lowY, move.x, rect.highY);
                block.blockKind = BlockKind::Cut;
                block.blockId = 0;
                block.parentBlockNum = blockNum;
            }
            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = Rect(move.x, rect.lowY, rect.highX, rect.highY);
                block.blockKind = BlockKind::Cut;
                block.blockId = 1;
                block.parentBlockNum = blockNum;
            }
            cost = calcCost(problem, problem.hasSourceImage ? 2 : 7, rect);
            break;
        }
        case MoveKind::CutY:
        {
            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = Rect(rect.lowX, rect.lowY, rect.highX, move.y);;
                block.blockKind = BlockKind::Cut;
                block.blockId = 0;
                block.parentBlockNum = blockNum;
            }
            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = Rect(rect.lowX, move.y, rect.highX, rect.highY);
                block.blockKind = BlockKind::Cut;
                block.blockId = 1;
                block.parentBlockNum = blockNum;
            }
            cost = calcCost(problem, problem.hasSourceImage ? 2 : 7, rect);
            break;
        }
        case MoveKind::CutXY:
        {
            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = Rect(rect.lowX, rect.lowY, move.x, move.y);
                block.blockKind = BlockKind::Cut;
                block.blockId = 0;
                block.parentBlockNum = blockNum;
            }
            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = Rect(move.x, rect.lowY, rect.highX, move.y);
                block.blockKind = BlockKind::Cut;
                block.blockId = 1;
                block.parentBlockNum = blockNum;
            }
            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = Rect(move.x, move.y, rect.highX, rect.highY);
                block.blockKind = BlockKind::Cut;
                block.blockId = 2;
                block.parentBlockNum = blockNum;
            }
            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = Rect(rect.lowX, move.y, move.x, rect.highY);
                block.blockKind = BlockKind::Cut;
                block.blockId = 3;
                block.parentBlockNum = blockNum;
            }
            cost = calcCost(problem, problem.hasSourceImage ? 3 : 10, rect);
            break;
        }
        default: ;
        }

        break;
    }
    case MoveKind::Swap:
    case MoveKind::Merge:
    {
        uint64_t blockNum0 = move.blockNums[0];
        if (blockNum0 >= canvas.blocks.size() ||
            !canvas.blocks[blockNum0].valid)
        {
            if (pMsg) *pMsg = strprintf("Move %" PRId64 ": Invalid block number\n", moveIdx);
            return false;
        }

        Rect rect0 = canvas.blocks[blockNum0].rect;

        uint64_t blockNum1 = move.blockNums[1];
        if (blockNum1 >= canvas.blocks.size() ||
            !canvas.blocks[blockNum1].valid)
        {
            if (pMsg) *pMsg = strprintf("Move %" PRId64 ": Invalid block number\n", moveIdx);
            return false;
        }

        Rect rect1 = canvas.blocks[blockNum1].rect;

        switch (move.moveKind)
        {
        case MoveKind::Swap:
        {
            uint64_t width = rect0.highX - rect0.lowX;
            uint64_t height = rect0.highY - rect0.lowY;
            if (rect1.highX - rect1.lowX != width ||
                rect1.highY - rect1.lowY != height)
            {
                if (pMsg) *pMsg = strprintf("Move %" PRId64 ": Swap size mismatch\n", moveIdx);
                return false;
            }

            if (pResult)
            {
                for (uint64_t y = 0; y < height; y++)
                {
                    for (uint64_t x = 0; x < width; x++)
                    {
                        using std::swap;
                        swap((*pResult)(rect0.lowX + x, rect0.lowY + y),
                             (*pResult)(rect1.lowX + x, rect1.lowY + y));
                    }
                }
            }

            if (pMoveIdxs)
            {
                for (uint64_t y = 0; y < height; y++)
                {
                    for (uint64_t x = 0; x < width; x++)
                    {
                        using std::swap;
                        swap((*pMoveIdxs)(rect0.lowX + x, rect0.lowY + y),
                             (*pMoveIdxs)(rect1.lowX + x, rect1.lowY + y));
                    }
                }
            }

            {
                using std::swap;
                swap(canvas.blocks[blockNum0].rect,
                     canvas.blocks[blockNum1].rect);
            }

            cost = calcCost(problem, 3, rect0);

            break;
        }
        case MoveKind::Merge:
        {
            canvas.blocks[blockNum0].valid = false;
            canvas.blocks[blockNum1].valid = false;

            Rect mergeRect;
            if (rect0.lowY == rect1.lowY &&
                rect0.highY == rect1.highY &&
                rect0.highX == rect1.lowX)
            {
                mergeRect = Rect(rect0.lowX,
                                 rect0.lowY,
                                 rect1.highX,
                                 rect0.highY);
            }
            else if (rect0.lowY == rect1.lowY &&
                     rect0.highY == rect1.highY &&
                     rect1.highX == rect0.lowX)
            {
                mergeRect = Rect(rect1.lowX,
                                 rect0.lowY,
                                 rect0.highX,
                                 rect0.highY);
            }
            else if (rect0.lowX == rect1.lowX &&
                     rect0.highX == rect1.highX &&
                     rect0.highY == rect1.lowY)
            {
                mergeRect = Rect(rect0.lowX,
                                 rect0.lowY,
                                 rect0.highX,
                                 rect1.highY);
            }
            else if (rect0.lowX == rect1.lowX &&
                     rect0.highX == rect1.highX &&
                     rect1.highY == rect0.lowY)
            {
                mergeRect = Rect(rect0.lowX,
                                 rect1.lowY,
                                 rect0.highX,
                                 rect0.highY);
            }
            else
            {
                if (pMsg) *pMsg = strprintf("Move %" PRId64 ": Merge blocks not compatible\n", moveIdx);
                return false;
            }

            {
                auto& block = canvas.blocks.emplace_back();
                block.rect = mergeRect;
                block.blockKind = BlockKind::Simple;
                block.blockId = canvas.nextBlockId++;
            }

            uint64_t rect0Size = rect0.size();
            uint64_t rect1Size = rect1.size();

            cost = calcCost(problem, 1, rect0Size >= rect1Size ? rect0 : rect1);

            break;
        }
        default: ;
        }

        break;
    }
    }

    if (pCost) *pCost = cost;
    return true;
}

bool eval(const Problem& problem,
          const Canvas& initialCanvas,
          const Image& initialImage,
          const vector<Move>& solution,
          Canvas& canvas,
          Image* pResult,
          uint64_t* pCost,
          Grid<uint64_t>* pMoveIdxs,
          vector<Rect>* pRects,
          string* pMsg)
{
    canvas = initialCanvas;

    if (pResult)
    {
        *pResult = initialImage;
    }

    if (pMoveIdxs)
    {
        uint64_t width = problem.width;
        uint64_t height = problem.height;
        pMoveIdxs->resize(width, height);
        for (uint64_t y = 0; y < height; y++)
        {
            for (uint64_t x = 0; x < width; x++)
            {
                (*pMoveIdxs)(x, y) = (uint64_t)-1;
            }
        }
    }

    if (pRects)
    {
        pRects->clear();
    }

    uint64_t cost = 0;

    uint64_t moveCount = solution.size();
    for (uint64_t moveIdx = 0; moveIdx < moveCount; moveIdx++)
    {
        const Move& move = solution[moveIdx];

        uint64_t curCost = 0;
        if (!evalMove(problem,
                      move,
                      canvas,
                      pResult,
                      &curCost,
                      pMoveIdxs,
                      moveIdx,
                      pRects,
                      pMsg))
        {
            return false;
        }

        cost += curCost;
    }

    if (pCost) *pCost = cost;
    return true;
}

bool eval(const Problem& problem,
          const vector<Move>& solution,
          Canvas& canvas,
          Image* pResult,
          uint64_t* pCost,
          Grid<uint64_t>* pMoveIdxs,
          vector<Rect>* pRects,
          string* pMsg)
{
    return eval(problem,
                problem.initialCanvas,
                problem.initialImage,
                solution,
                canvas,
                pResult,
                pCost,
                pMoveIdxs,
                pRects,
                pMsg);
}
