#include "FindColor.hpp"
#include <cmath>

using std::vector;

#define USE_CACHE 1

namespace
{
    double pixelDiff(Color p1, Color p2)
    {
        uint32_t r = (uint32_t)p1.r - (uint32_t)p2.r;
        uint32_t g = (uint32_t)p1.g - (uint32_t)p2.g;
        uint32_t b = (uint32_t)p1.b - (uint32_t)p2.b;

        r *= r;
        g *= g;
        b *= b;

        return std::sqrt(r + g + b);
    }

    double calcDiff(const Image& targetImage,
                    const vector<Coord>& coords,
                    Color color,
                    uint64_t stride)
    {
        double diff = 0.0;
        uint64_t coordCount = coords.size();
        for (uint64_t coordIdx = 0; coordIdx < coordCount; coordIdx += stride)
        {
            Coord coord = coords[coordIdx];
            diff += pixelDiff(targetImage(coord.x, coord.y), color);
        }
        return diff;
    }
}

Color findColor(const Problem& problem,
                const vector<Coord>& coords)
{
    if (coords.empty())
    {
        return Color(255, 255, 255, 255);

    }

#if USE_CACHE
    {
        Color color;
        if (problem.findColorCache.lookup(coords, &color))
        {
            return color;
        }
    }
#endif

    Color color(128, 128, 128, 255);

    double diff = 0.0;
    uint64_t prevStride = 0;
    uint32_t didNothingCount = 0;

    auto checkOne = [&](uint8_t deltaR,
                        uint8_t deltaG,
                        uint8_t deltaB,
                        uint64_t stride) -> bool
    {
        if (stride != prevStride)
        {
            prevStride = stride;
            diff = calcDiff(problem.targetImage, coords, color, stride);
        }

        Color curColor(color.r + deltaR,
                       color.g + deltaG,
                       color.b + deltaB,
                       255);
        double curDiff = calcDiff(problem.targetImage, coords, curColor, stride);
        if (curDiff < diff)
        {
            color = curColor;
            diff = curDiff;
            return true;
        }

        return false;
    };

    auto checkTwo = [&](uint8_t deltaR,
                        uint8_t deltaG,
                        uint8_t deltaB,
                        uint64_t stride) -> bool
    {
        if (checkOne(deltaR, deltaG, deltaB, stride) ||
            checkOne(-deltaR, -deltaG, -deltaB, stride))
        {
            didNothingCount = 0;
        }
        else
        {
            didNothingCount++;
        }
#if 0
        uint64_t curCoordCount = coords.size() / stride;
        printf("%3" PRIu8 " %3" PRIu8 " %3" PRIu8 "  %" PRIu64 "\n", color.r, color.g, color.b, curCoordCount);
#endif
        return didNothingCount >= 3;
    };

#if 0
    auto check = [&](uint8_t delta,
                     uint64_t stride)
    {
        checkTwo(delta, 0, 0, stride);
        checkTwo(0, delta, 0, stride);
        checkTwo(0, 0, delta, stride);
    };
#endif

    auto checkFull = [&](uint8_t delta,
                         uint64_t stride)
    {
        didNothingCount = 0;
        while (true)
        {
            if (checkTwo(delta, 0, 0, stride)) break;
            if (checkTwo(0, delta, 0, stride)) break;
            if (checkTwo(0, 0, delta, stride)) break;
        }
    };

    checkFull(64, 1024);
    checkFull(32, 512);
    checkFull(16, 256);
    checkFull(8, 128);
    checkFull(4, 64);
    checkFull(2, 32);
    checkFull(1, 16);
    checkFull(1, 1);

#if 0
    printf("\n");
#endif

#if USE_CACHE
    problem.findColorCache.add(coords, color);
#endif

    return color;
}
