#include "Problem.hpp"
#include "StbImage.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include "Cleanup.hpp"
#include "Json.hpp"
#include "JsonUtils.hpp"

using std::string;
using std::vector;
using std::unique_ptr;

namespace J = rapidjson;

namespace
{
    class InitialBlock
    {
    public:
        string blockId;
        Block block;
        Color color;
    };

    void parsePoint(const J::Value& jPoint,
                    uint64_t* pX,
                    uint64_t* pY)
    {
        auto& x = *pX;
        auto& y = *pY;

        JCheckArray(jPoint);

        if (jPoint.Size() != 2)
        {
            throw JException("jPoint: incorrect size");
        }

        {
            auto& jX = jPoint[0];
            x = JGetUint64(jX);
        }

        {
            auto& jY = jPoint[1];
            y = JGetUint64(jY);
        }
    }

    void parseComponent(const J::Value& jComponent,
                        uint8_t* pComponent)
    {
        auto& component = *pComponent;

        uint64_t component64 = JGetUint64(jComponent);
        if (component64 > 255)
        {
            throw JException("jComponent: color component too large");
        }

        component = component64;
    }

    void parseColor(const J::Value& jColor,
                    Color* pColor)
    {
        auto& color = *pColor;

        JCheckArray(jColor);

        if (jColor.Size() != 4)
        {
            throw JException("jColor: incorrect size");
        }

        {
            auto& jR = jColor[0];
            parseComponent(jR, &color.r);
        }

        {
            auto& jG = jColor[1];
            parseComponent(jG, &color.g);
        }

        {
            auto& jB = jColor[2];
            parseComponent(jB, &color.b);
        }

        {
            auto& jA = jColor[3];
            parseComponent(jA, &color.a);
            if (color.a != 255)
            {
                throw JException("jA: color alpha not 255");
            }
        }
    }

    void parseBlock(const J::Value& jBlock,
                    Problem& problem,
                    const Image* pSourceImage)
    {
        JCheckObject(jBlock);

        string blockId = JObjectGetString(jBlock, "blockId");

        uint64_t blockNum = problem.initialCanvas.blocks.size();
        if (blockId != strprintf("%" PRIu64 "", blockNum))
        {
            throw JException("jBlock: incorrect block id");
        }

        Rect rect;

        {
            auto& jBottomLeft = JObjectGetArray(jBlock, "bottomLeft");
            parsePoint(jBottomLeft, &rect.lowX, &rect.lowY);
        }

        {
            auto& jTopRight = JObjectGetArray(jBlock, "topRight");
            parsePoint(jTopRight, &rect.highX, &rect.highY);
        }

        auto& block = problem.initialCanvas.blocks.emplace_back();
        block.rect = rect;
        block.blockKind = BlockKind::Simple;
        block.blockId = blockNum;

        if (jBlock.HasMember("color"))
        {
            Color color;

            {
                auto& jColor = JObjectGetArray(jBlock, "color");
                parseColor(jColor, &color);
            }

            for (uint64_t y = rect.lowY; y < rect.highY; y++)
            {
                for (uint64_t x = rect.lowX; x < rect.highX; x++)
                {
                    problem.initialImage(x, y) = color;
                }
            }
        }
        else if (jBlock.HasMember("pngBottomLeftPoint"))
        {
            if (!pSourceImage)
            {
                throw JException("jBlock: no source image");
            }

            uint64_t sourceX = 0;
            uint64_t sourceY = 0;

            {
                auto& jPngBottomLeftPoint = JObjectGetArray(jBlock, "pngBottomLeftPoint");
                parsePoint(jPngBottomLeftPoint, &sourceX, &sourceY);
            }

            for (uint64_t y = rect.lowY; y < rect.highY; y++)
            {
                for (uint64_t x = rect.lowX; x < rect.highX; x++)
                {
                    problem.initialImage(x, y) = (*pSourceImage)(sourceX + x, sourceY + y);
                }
            }
        }
        else
        {
            throw JException("jBlock: expected color or pngBottomLeftPoint");
        }
    }

    void parseBlocks(const J::Value& jBlocks,
                     Problem& problem,
                     const Image* pSourceImage)
    {
        JCheckArray(jBlocks);

        for (auto& jBlock : jBlocks.GetArray())
        {
            parseBlock(jBlock, problem, pSourceImage);
        }
    }

    void parseCanvas(const J::Value& jCanvas,
                     Problem& problem,
                     const Image* pSourceImage)
    {
        JCheckObject(jCanvas);

        uint64_t width = JObjectGetUint64(jCanvas, "width");
        uint64_t height = JObjectGetUint64(jCanvas, "height");

        if (width != problem.width ||
            height != problem.height)
        {
            throw JException("jCanvas: canvas size mismatch");
        }

        {
            auto& jBlocks = JObjectGetArray(jCanvas, "blocks");
            problem.initialCanvas.nextBlockId = jBlocks.Size();
            parseBlocks(jBlocks, problem, pSourceImage);
        }
    }

    bool parseCanvas(const string& canvasData,
                     Problem& problem,
                     const Image* pSourceImage,
                     string* pMsg)
    {
        try
        {
            J::Document jCanvas;
            jCanvas.Parse(canvasData);
            JCheckParseObject(jCanvas);

            parseCanvas(jCanvas, problem, pSourceImage);
        }
        catch (const JException& e)
        {
            if (pMsg) *pMsg = e.m_msg;
            return false;
        }

        return true;
    }

    bool parseImage(const vector<uint8_t>& imageData,
                    Image* pImage,
                    string* pMsg = nullptr)
    {
        int intWidth;
        int intHeight;
        int channelsInFile;
        unsigned char* data = stbi_load_from_memory((const unsigned char*)imageData.data(), imageData.size(), &intWidth, &intHeight, &channelsInFile, 4);

        if (!data)
        {
            if (pMsg) *pMsg = "Error parsing image";
            return false;
        }

        Cleanup freeCleanup([&] { free(data); });

        if (intWidth != 400 || intHeight != 400)
        {
            if (pMsg) *pMsg = "Unexpected image dimensions";
            return false;
        }

        if (channelsInFile != 4)
        {
            if (pMsg) *pMsg = "Unexpected number of channels in image";
            return false;
        }

        uint64_t width = intWidth;
        uint64_t height = intHeight;

        pImage->resize(width, height);
        for (uint64_t y = 0; y < height; y++)
        {
            for (uint64_t x = 0; x < width; x++)
            {
                uint64_t idx = (height - 1 - y) * width + x;
#if 1
                uint32_t alpha = (uint32_t)data[idx * 4 + 3];
                if (alpha != 255)
                {
                    if (pMsg) *pMsg = "Unexpected alpha value";
                    return false;
                }
#endif
                (*pImage)(x, y) = Color(data[idx * 4 + 0],
                                        data[idx * 4 + 1],
                                        data[idx * 4 + 2],
                                        data[idx * 4 + 3]);
            }
        }

        return true;
    }
}

bool loadProblem(uint64_t problemId, Problem* pProblem, string* pMsg)
{
    Problem problem;

    {
        string targetImageFileName = strprintf("../problems/%" PRIu64 ".png", problemId);
        vector<uint8_t> targetImageData;
        if (!readFile(targetImageFileName, &targetImageData))
        {
            if (pMsg) *pMsg = "Error reading target image file";
            return false;
        }

        if (!parseImage(targetImageData, &problem.targetImage, pMsg))
        {
            return false;
        }
    }

    uint64_t width = problem.targetImage.getWidth();
    uint64_t height = problem.targetImage.getHeight();

    problem.width = width;
    problem.height = height;

    unique_ptr<Image> pSourceImage;

    {
        string sourceImageFileName = strprintf("../problems/%" PRIu64 ".source.png", problemId);
        vector<uint8_t> sourceImageData;
        if (readFile(sourceImageFileName, &sourceImageData))
        {
            pSourceImage.reset(new Image);
            if (!parseImage(sourceImageData, pSourceImage.get(), pMsg))
            {
                return false;
            }

            problem.hasSourceImage = true;
        }
    }

    problem.initialImage.resize(width, height);

    string canvasFileName = strprintf("../problems/%" PRIu64 ".initial.json", problemId);
    string canvasData;
    if (!readFile(canvasFileName, &canvasData))
    {
        auto& block = problem.initialCanvas.blocks.emplace_back();
        block.rect = Rect(0, 0, width, height);
        block.blockKind = BlockKind::Simple;
        block.blockId = 0;

        problem.initialCanvas.nextBlockId = 1;
    }
    else
    {
        if (!parseCanvas(canvasData, problem, pSourceImage.get(), pMsg))
        {
            return false;
        }
    }

    if (pProblem) *pProblem = std::move(problem);
    return true;
}
