#include "Scores.hpp"
#include "FileUtils.hpp"
#include "Json.hpp"
#include "JsonUtils.hpp"

using std::string;
using std::vector;

namespace J = rapidjson;

bool readSubmissions(string* pSubmissionsData)
{
    string fileName = "../data/submissions";
    return readFile(fileName, pSubmissionsData);
}

bool writeSubmissions(const string& submissionsData)
{
    string fileName = "../data/submissions";
    return writeFile(fileName, submissionsData);
}

bool readResults(string* pResultsData)
{
    string fileName = "../data/results";
    return readFile(fileName, pResultsData);
}

bool writeResults(const string& resultsData)
{
    string fileName = "../data/results";
    return writeFile(fileName, resultsData);
}

namespace
{
    void parseSubmission(const J::Value& jSubmission,
                         Submission* pSubmission)
    {
        auto& submission = *pSubmission;

        JCheckObject(jSubmission);

        submission.submissionId = JObjectGetUint64(jSubmission, "id");
        submission.id = JObjectGetUint64(jSubmission, "problem_id");
        submission.submittedAt = JObjectGetString(jSubmission, "submitted_at");
        submission.status = JObjectGetString(jSubmission, "status");
        submission.score = JObjectGetUint64(jSubmission, "score");
    }

    void parseSubmissions(const J::Value& jSubmissions,
                          vector<Submission>* pSubmissions)
    {
        auto& submissions = *pSubmissions;

        JCheckArray(jSubmissions);

        submissions.clear();
        submissions.reserve(jSubmissions.Size());
        for (auto& jSubmission : jSubmissions.GetArray())
        {
            auto& submission = submissions.emplace_back();
            parseSubmission(jSubmission, &submission);
        }
    }

    void parseSubmissionsFile(const J::Value& jSubmissionsFile,
                              vector<Submission>* pSubmissions)
    {
        JCheckObject(jSubmissionsFile);

        {
            auto& jSubmissions = JObjectGetArray(jSubmissionsFile, "submissions");
            parseSubmissions(jSubmissions, pSubmissions);
        }
    }

}

bool parseSubmissions(const string& submissionsData,
                      vector<Submission>* pSubmissions,
                      string* pMsg)
{
    try
    {
        J::Document jSubmissionsFile;
        jSubmissionsFile.Parse(submissionsData);
        JCheckParseObject(jSubmissionsFile);

        parseSubmissionsFile(jSubmissionsFile, pSubmissions);
    }
    catch (const JException& e)
    {
        if (pMsg) *pMsg = e.m_msg;
        return false;
    }

    for (auto& submission : *pSubmissions)
    {
        if (submission.status != "SUCCEEDED")
        {
            if (pMsg) *pMsg = "Submission status is '" + submission.status + "'";
            return false;
        }

        if (submission.score == 0)
        {
            if (pMsg) *pMsg = "Submission score is 0";
            return false;
        }
    }

    return true;
}

bool parseSubmissions(vector<Submission>* pSubmissions,
                      string* pMsg)
{
    string submissionsData;
    if (!readSubmissions(&submissionsData))
    {
        if (pMsg) *pMsg = "Error reading submissions";
        return false;
    }

    return parseSubmissions(submissionsData, pSubmissions, pMsg);
}

void getMine(const vector<Submission>& submissions,
             vector<uint64_t>* pMine)
{
    vector<uint64_t> mine;
    for (auto& submission : submissions)
    {
        if (submission.id > 1000000) continue;
        if (mine.size() < submission.id + 1)
        {
            mine.resize(submission.id + 1, (uint64_t)-1);
        }
        if (submission.score < mine[submission.id])
        {
            mine[submission.id] = submission.score;
        }
    }

    if (pMine) *pMine = std::move(mine);
}

namespace
{
    void parseResult(const J::Value& jResult,
                     Result* pResult)
    {
        auto& result = *pResult;

        JCheckObject(jResult);

        result.id = JObjectGetUint64(jResult, "problem_id");
        result.problemName = JObjectGetString(jResult, "problem_name");
        result.lastSubmittedAt = JObjectGetString(jResult, "last_submitted_at");
        result.submissionCount = JObjectGetUint64(jResult, "submission_count");
        result.minCost = JObjectGetUint64(jResult, "min_cost");
        result.overallBestCost = JObjectGetUint64(jResult, "overall_best_cost");
    }

    void parseResults(const J::Value& jResults,
                      vector<Result>* pResults)
    {
        auto& results = *pResults;

        JCheckArray(jResults);

        results.clear();
        results.reserve(jResults.Size());
        for (auto& jResult : jResults.GetArray())
        {
            auto& result = results.emplace_back();
            parseResult(jResult, &result);
        }
    }

    void parseResultsFile(const J::Value& jResultsFile,
                          Results* pResults)
    {
        auto& results = *pResults;

        JCheckObject(jResultsFile);

        {
            auto& jResults = JObjectGetArray(jResultsFile, "results");
            parseResults(jResults, &results.results);
        }

        results.totalCost = JObjectGetUint64(jResultsFile, "total_cost");
        results.solvedProblemCount = JObjectGetUint64(jResultsFile, "solved_problem_count");
    }

}

bool parseResults(const string& resultsData,
                  Results* pResults,
                  string* pMsg)
{
    try
    {
        J::Document jResultsFile;
        jResultsFile.Parse(resultsData);
        JCheckParseObject(jResultsFile);

        parseResultsFile(jResultsFile, pResults);
    }
    catch (const JException& e)
    {
        if (pMsg) *pMsg = e.m_msg;
        return false;
    }

    return true;
}

bool parseResults(Results* pResults,
                  string* pMsg)
{
    string resultsData;
    if (!readResults(&resultsData))
    {
        if (pMsg) *pMsg = "Error reading results";
        return false;
    }

    return parseResults(resultsData, pResults, pMsg);
}

void getMine(const Results& results,
             vector<uint64_t>* pMine)
{
    vector<uint64_t> mine;
    for (auto& result : results.results)
    {
        if (result.id > 1000000) continue;
        if (result.submissionCount == 0) continue;
        if (result.minCost == 0) continue;
        if (mine.size() < result.id + 1)
        {
            mine.resize(result.id + 1, (uint64_t)-1);
        }
        if (result.minCost < mine[result.id])
        {
            mine[result.id] = result.minCost;
        }
    }

    if (pMine) *pMine = std::move(mine);
}

void getBest(const Results& results,
             vector<uint64_t>* pBest)
{
    vector<uint64_t> best;
    for (auto& result : results.results)
    {
        if (result.id > 1000000) continue;
        if (result.overallBestCost == 0) continue;
        if (best.size() < result.id + 1)
        {
            best.resize(result.id + 1, (uint64_t)-1);
        }
        if (result.overallBestCost < best[result.id])
        {
            best[result.id] = result.overallBestCost;
        }
    }

    if (pBest) *pBest = std::move(best);
}
