#include "Solution.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include "Problem.hpp"
#include "Canvas.hpp"
#include "Eval.hpp"
#include <limits>

using std::string;
using std::vector;

string formatBlockNum(const Canvas& canvas, uint64_t blockNum)
{
    string str("]");
    while (true)
    {
        auto& block = canvas.blocks[blockNum];
        if (block.blockKind == BlockKind::Simple)
        {
            str = strprintf("[%" PRIu64 "", block.blockId) + str;
            break;
        }
        else
        {
            str = strprintf(".%" PRIu64 "", block.blockId) + str;
            blockNum = block.parentBlockNum;
        }
    }
    return str;
}

string formatColor(const Color& color)
{
    return strprintf("[%" PRIu8 ", %" PRIu8 ", %" PRIu8 ", %" PRIu8 "]",
                     color.r,
                     color.g,
                     color.b,
                     color.a);
}

string formatMove(const Canvas& canvas, const Move& move)
{
    switch (move.moveKind)
    {
    case MoveKind::Color:
        return strprintf("color %s %s",
                         formatBlockNum(canvas, move.blockNums[0]).c_str(),
                         formatColor(move.color).c_str());
    case MoveKind::CutX:
        return strprintf("cut %s [x] [%" PRIu64 "]",
                         formatBlockNum(canvas, move.blockNums[0]).c_str(),
                         move.x);
    case MoveKind::CutY:
        return strprintf("cut %s [y] [%" PRIu64 "]",
                         formatBlockNum(canvas, move.blockNums[0]).c_str(),
                         move.y);
    case MoveKind::CutXY:
        return strprintf("cut %s [%" PRIu64 ", %" PRIu64 "]",
                         formatBlockNum(canvas, move.blockNums[0]).c_str(),
                         move.x,
                         move.y);
    case MoveKind::Swap:
        return strprintf("swap %s %s",
                         formatBlockNum(canvas, move.blockNums[0]).c_str(),
                         formatBlockNum(canvas, move.blockNums[1]).c_str());
    case MoveKind::Merge:
        return strprintf("merge %s %s",
                         formatBlockNum(canvas, move.blockNums[0]).c_str(),
                         formatBlockNum(canvas, move.blockNums[1]).c_str());
    }

    fprintf(stderr, "formatMove: Unhandled move kind\n");
    exit(1);
    return string();
}

string formatSolution(const Problem& problem,
                      const Solution& solution)
{
    Canvas canvas(problem.initialCanvas);
    string msg;

    string str;
    for (auto& move : solution)
    {
        if (!evalMove(problem,
                      move,
                      canvas,
                      nullptr,
                      nullptr,
                      nullptr,
                      (uint64_t)-1,
                      nullptr,
                      &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        str += formatMove(canvas, move);
        str += '\n';
    }
    return str;
}

namespace
{
    void skipWhitespace(const string& str, size_t& pos)
    {
        size_t size = str.size();
        while (pos < size)
        {
            char ch = str[pos];
            if (ch == ' ' ||
                ch == '\t')
            {
                pos++;
            }
            else
            {
                break;
            }
        }
    }

    bool parseChar(const string& str, size_t& pos, char ch)
    {
        size_t size = str.size();
        if (pos < size && str[pos] == ch)
        {
            pos++;
            return true;
        }
        return false;
    }

    bool parseStr(const string& str, size_t& pos, const string& toMatch)
    {
        //size_t size = str.size();
        if (str.starts_with(toMatch))
        {
            pos += toMatch.size();
            return true;
        }
        return false;
    }

    bool parseBlockId(const string& str, size_t& pos, const Canvas& canvas, const std::unordered_map<string, uint64_t>& blockMap, string* pBlockId, uint64_t* pBlockNum)
    {
        size_t size = str.size();
        string blockId;
        if (!parseChar(str, pos, '[')) return false;
        skipWhitespace(str, pos);
        while (pos < size && str[pos] != ']' && str[pos] != ' ' && str[pos] != '\t')
        {
            blockId += str[pos];
            pos++;
        }
        skipWhitespace(str, pos);
        if (!parseChar(str, pos, ']')) return false;
        auto it = blockMap.find(blockId);
        if (it == blockMap.end()) return false;
        uint64_t blockNum = it->second;
        if (blockNum >= canvas.blocks.size()) return false;
        if (!canvas.blocks[blockNum].valid) return false;
        *pBlockId = blockId;
        *pBlockNum = blockNum;
        return true;
    }

    bool parseNum(const string& str, size_t& pos, uint64_t* pNum)
    {
        size_t size = str.size();
        size_t curPos = pos;
        if (curPos >= size) return false;
        if (str[curPos] < '0' || str[curPos] > '9') return false;
        uint64_t num = 0;
        while (true)
        {
            if (curPos >= size) break;
            char ch = str[curPos];
            if (ch < '0' || ch > '9') break;
            curPos++;
            if (num > std::numeric_limits<uint64_t>::max() / 10) return false;
            num *= 10;
            uint64_t digit = ch - '0';
            if (digit > std::numeric_limits<uint64_t>::max() - num) return false;
            num += digit;
        }
        pos = curPos;
        *pNum = num;
        return true;
    }

    bool parseColor(const string& str, size_t& pos, Color* pColor)
    {
        //size_t size = str.size();
        if (!parseChar(str, pos, '[')) return false;
        skipWhitespace(str, pos);
        uint64_t r;
        if (!parseNum(str, pos, &r)) return false;
        skipWhitespace(str, pos);
        if (!parseChar(str, pos, ',')) return false;
        skipWhitespace(str, pos);
        uint64_t g;
        if (!parseNum(str, pos, &g)) return false;
        skipWhitespace(str, pos);
        if (!parseChar(str, pos, ',')) return false;
        skipWhitespace(str, pos);
        uint64_t b;
        if (!parseNum(str, pos, &b)) return false;
        skipWhitespace(str, pos);
        if (!parseChar(str, pos, ',')) return false;
        skipWhitespace(str, pos);
        uint64_t a;
        if (!parseNum(str, pos, &a)) return false;
        skipWhitespace(str, pos);
        if (!parseChar(str, pos, ']')) return false;
        if (r > 255 || g > 255 || b > 255 || a > 255) return false;
        *pColor = Color(r, g, b, a);
        return true;
    }
}

bool parseSolution(const Problem& problem,
                   const string& strIn,
                   Solution* pSolution)
{
    vector<string> lines;
    getLines(strIn, &lines);

    std::unordered_map<string, uint64_t> blockMap;
    {
        uint64_t blockCount = problem.initialCanvas.blocks.size();
        for (uint64_t blockNum = 0; blockNum < blockCount; blockNum++)
        {
            auto& block = problem.initialCanvas.blocks[blockNum];
            blockMap[strprintf("%" PRIu64 "", block.blockId)] = blockNum;
        }
    }

    Canvas canvas(problem.initialCanvas);

    Solution solution;

    string msg;

    for (auto& str : lines)
    {
        size_t pos = 0;
        size_t size = str.size();
        skipWhitespace(str, pos);
        if (pos == size) continue;
        if (str[pos] == '#') continue;

        Move& move = solution.emplace_back();

        if (parseStr(str, pos, "color"))
        {
            move.moveKind = MoveKind::Color;
            skipWhitespace(str, pos);
            string blockId0;
            if (!parseBlockId(str, pos, canvas, blockMap, &blockId0, &move.blockNums[0])) return false;
            skipWhitespace(str, pos);
            if (!parseColor(str, pos, &move.color)) return false;
        }
        else if (parseStr(str, pos, "cut"))
        {
            skipWhitespace(str, pos);
            string blockId0;
            if (!parseBlockId(str, pos, canvas, blockMap, &blockId0, &move.blockNums[0])) return false;
            skipWhitespace(str, pos);
            if (!parseChar(str, pos, '[')) return false;
            skipWhitespace(str, pos);
            if (parseChar(str, pos, 'x') || parseChar(str, pos, 'X'))
            {
                move.moveKind = MoveKind::CutX;
                skipWhitespace(str, pos);
                if (!parseChar(str, pos, ']')) return false;
                skipWhitespace(str, pos);
                if (!parseChar(str, pos, '[')) return false;
                skipWhitespace(str, pos);
                if (!parseNum(str, pos, &move.x)) return false;
                skipWhitespace(str, pos);
                if (!parseChar(str, pos, ']')) return false;
                blockMap[blockId0 + ".0"] = canvas.blocks.size() + 0;
                blockMap[blockId0 + ".1"] = canvas.blocks.size() + 1;
            }
            else if (parseChar(str, pos, 'y') || parseChar(str, pos, 'Y'))
            {
                move.moveKind = MoveKind::CutY;
                skipWhitespace(str, pos);
                if (!parseChar(str, pos, ']')) return false;
                skipWhitespace(str, pos);
                if (!parseChar(str, pos, '[')) return false;
                skipWhitespace(str, pos);
                if (!parseNum(str, pos, &move.y)) return false;
                skipWhitespace(str, pos);
                if (!parseChar(str, pos, ']')) return false;
                blockMap[blockId0 + ".0"] = canvas.blocks.size() + 0;
                blockMap[blockId0 + ".1"] = canvas.blocks.size() + 1;
            }
            else
            {
                move.moveKind = MoveKind::CutXY;
                if (!parseNum(str, pos, &move.x)) return false;
                skipWhitespace(str, pos);
                if (!parseChar(str, pos, ',')) return false;
                skipWhitespace(str, pos);
                if (!parseNum(str, pos, &move.y)) return false;
                skipWhitespace(str, pos);
                if (!parseChar(str, pos, ']')) return false;
                blockMap[blockId0 + ".0"] = canvas.blocks.size() + 0;
                blockMap[blockId0 + ".1"] = canvas.blocks.size() + 1;
                blockMap[blockId0 + ".2"] = canvas.blocks.size() + 2;
                blockMap[blockId0 + ".3"] = canvas.blocks.size() + 3;
            }
        }
        else if (parseStr(str, pos, "swap"))
        {
            move.moveKind = MoveKind::Swap;
            skipWhitespace(str, pos);
            string blockId0;
            if (!parseBlockId(str, pos, canvas, blockMap, &blockId0, &move.blockNums[0])) return false;
            skipWhitespace(str, pos);
            string blockId1;
            if (!parseBlockId(str, pos, canvas, blockMap, &blockId1, &move.blockNums[1])) return false;
        }
        else if (parseStr(str, pos, "merge"))
        {
            move.moveKind = MoveKind::Merge;
            skipWhitespace(str, pos);
            string blockId0;
            if (!parseBlockId(str, pos, canvas, blockMap, &blockId0, &move.blockNums[0])) return false;
            skipWhitespace(str, pos);
            string blockId1;
            if (!parseBlockId(str, pos, canvas, blockMap, &blockId1, &move.blockNums[1])) return false;
            blockMap[strprintf("%" PRIu64 "", canvas.nextBlockId)] = canvas.blocks.size() + 0;
        }
        else
        {
            return false;
        }

        if (!evalMove(problem,
                      move,
                      canvas,
                      nullptr,
                      nullptr,
                      nullptr,
                      (uint64_t)-1,
                      nullptr,
                      &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            return false;
        }

        skipWhitespace(str, pos);
        if (pos != size) return false;
    }

    if (pSolution) *pSolution = std::move(solution);
    return true;
}
