#include "Common.hpp"
#include "ParseUtils.hpp"
#include "ProblemCount.hpp"
#include "Problem.hpp"
#include "Best.hpp"
#include "CalcScore.hpp"
#include "Merge.hpp"
#include "Rects.hpp"
#include "FindColors.hpp"
#include "Eval.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: buddy [<options>]\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
}

int main(int argc, char *argv[])
{
    bool help = false;
    bool saveChanges = true;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveChanges = false;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    string msg;

    for (uint64_t toId = 1; toId <= problemCount; toId++)
    {
        Problem toProblem;
        if (!loadProblem(toId, &toProblem))
        {
            printf("Error loading problem\n");
            return 1;
        }

        Best toBest;
        toBest.load(toId);

        uint64_t bestScore = toBest.getScore();

        uint64_t width = toProblem.width;
        uint64_t height = toProblem.height;

        for (uint64_t fromId = 1; fromId <= problemCount; fromId++)
        {
            if (fromId == toId) continue;

            Problem fromProblem;
            if (!loadProblem(fromId, &fromProblem))
            {
                printf("Error loading problem\n");
                return 1;
            }

            if (fromProblem.width != width ||
                fromProblem.height != height)
            {
                continue;
            }

            bool different = false;
            for (uint64_t y = 0; y < height; y++)
            {
                for (uint64_t x = 0; x < width; x++)
                {
                    if (fromProblem.targetImage(x, y) != toProblem.targetImage(x, y))
                    {
                        different = true;
                        break;
                    }
                }
                if (different)
                {
                    break;
                }
            }
            if (different)
            {
                continue;
            }

            Best fromBest;
            fromBest.load(fromId);

            Canvas canvas(toProblem.initialCanvas);
            Solution solution;

            if (!merge(toProblem, canvas, solution))
            {
                printf("Merge failed\n");
                return 1;
            }

            size_t mergeSolutionSize = solution.size();

            vector<Rect> rects;

            {
                Canvas tmpCanvas;
                if (!eval(fromProblem, fromBest.getSolution(), tmpCanvas, nullptr, nullptr, nullptr, &rects, &msg))
                {
                    printf("%s\n", msg.c_str());
                    return 1;
                }
            }

            {
                auto& move = solution.emplace_back();
                move.moveKind = MoveKind::Color;
                move.blockNums[0] = canvas.blocks.size() - 1;
                //move.color = ...;
            }

            makeRects(toProblem, rects, canvas, solution);

            findColors(toProblem, solution);

            if (toProblem.initialCanvas.blocks.size() == 1 &&
                !toProblem.hasSourceImage &&
                solution[mergeSolutionSize].color == toProblem.initialImage(0, 0))
            {
                solution.erase(solution.begin() + mergeSolutionSize);
            }

            uint64_t score = 0;
            if (!calcScore(toProblem, solution, nullptr, &score, nullptr, nullptr, nullptr, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }

            bool better = score < bestScore;
            bool same = score == bestScore;
            bool newBest = better && saveChanges && toBest.update(solution);

            printf("%2" PRIu64 "  %2" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "%s\n",
                   toId,
                   fromId,
                   bestScore,
                   score,
                   newBest ? "  BEST" :
                   better ? "  best" :
                   same ? "  same" : "");

            if (score < bestScore)
            {
                bestScore = score;
            }
        }
    }

    return 0;
}
