#include "Common.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "TimeUtils.hpp"
#include "Graphics.hpp"
#include "ProblemCount.hpp"
#include "Problem.hpp"
#include "Best.hpp"
#include "CalcScore.hpp"
#include "Merge.hpp"
#include "Rects.hpp"
#include "FindColors.hpp"
#include "Eval.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: edit [<options>] <id>\n");
    printf("  <id>  Problem id\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
    printf("  -g <granularity>\n");
    printf("        Start with an initial grid of rectangles\n");
    printf("  -x    Flip grid in the X direction\n");
    printf("  -y    Flip grid in the Y direction\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;
    bool gotGrid = false;

    bool help = false;
    bool saveChanges = true;
    uint64_t id = 0;
    uint64_t gridSize = 0;
    bool flipX = false;
    bool flipY = false;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveChanges = false;
        }
        else if (strArg == "-g")
        {
            if (gotGrid || iArg >= argc)
            {
                usage();
                return 1;
            }
            strArg = argv[iArg++];
            if (!parseU64(strArg, &gridSize) || gridSize < 1 || gridSize > 400)
            {
                usage();
                return 1;
            }
            gotGrid = true;
        }
        else if (strArg == "-x")
        {
            flipX = true;
        }
        else if (strArg == "-y")
        {
            flipY = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > problemCount)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    Problem problem;

    string msg;
    if (!loadProblem(id, &problem, &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        return 1;
    }

    Best best;
    best.load(id);

    uint64_t bestScore = (uint64_t)-1;
    uint64_t bestCount = 0;

    Canvas canvas(problem.initialCanvas);
    Solution solution;

    if (!merge(problem, canvas, solution))
    {
        printf("Merge failed\n");
        return 1;
    }

    size_t mergeSolutionSize = solution.size();

    uint64_t width = problem.width;
    uint64_t height = problem.height;

    vector<Rect> rects;

    if (gotGrid)
    {
        for (uint64_t y = gridSize; y <= height; y += gridSize)
        {
            for (uint64_t x = gridSize; x <= width; x += gridSize)
            {
                if (flipX && flipY)
                {
                    rects.emplace_back(width - x, height - y, width, height);
                }
                else if (flipX)
                {
                    rects.emplace_back(width - x, 0, width, y);
                }
                else if (flipY)
                {
                    rects.emplace_back(0, height - y, x, height);
                }
                else
                {
                    rects.emplace_back(0, 0, x, y);
                }
            }
        }
        std::reverse(rects.begin(), rects.end());
    }

    vector<vector<Rect>> undoStack;
    vector<vector<Rect>> redoStack;

    auto beginTransaction = [&]()
    {
        undoStack.push_back(rects);
    };

    auto endTransaction = [&]()
    {
        if (rects != undoStack.back())
        {
            redoStack.clear();
        }
        else
        {
            undoStack.pop_back();
        }
    };

    Image result;
    uint64_t score = 0;
    uint64_t similarity = 0;
    uint64_t cost = 0;
    Image errorImage;

    auto check = [&](const vector<Rect>& curRects,
                     bool keepAlways,
                     bool keepIfSame = false) -> bool
    {
        {
            auto& move = solution.emplace_back();
            move.moveKind = MoveKind::Color;
            move.blockNums[0] = canvas.blocks.size() - 1;
            //move.color = ...;
        }

        Canvas curCanvas(canvas);
        makeRects(problem, curRects, curCanvas, solution);

        findColors(problem, solution);

        if (problem.initialCanvas.blocks.size() == 1 &&
            !problem.hasSourceImage &&
            solution[mergeSolutionSize].color == problem.initialImage(0, 0))
        {
            solution.erase(solution.begin() + mergeSolutionSize);
        }

        Image curResult;
        uint64_t curScore = 0;
        uint64_t curSimilarity = 0;
        uint64_t curCost = 0;
        Image curErrorImage;
        if (!calcScore(problem, solution, &curResult, &curScore, &curSimilarity, &curCost, &curErrorImage, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        bool keep = keepAlways || curScore < score || (curScore == score && keepIfSame);

        if (keep)
        {
            rects = curRects;
            result = std::move(curResult);
            score = curScore;
            similarity = curSimilarity;
            cost = curCost;
            errorImage = std::move(curErrorImage);

            bool better = score < bestScore;
            bool same = score == bestScore;
            bool newBest = better && saveChanges && best.update(solution);

            if (better)
            {
                bestScore = score;
                bestCount = 1;
            }
            else if (same)
            {
                bestCount++;
            }

            if (better /*|| same*/)
            {
                printf("%2" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "%s\n",
                       id,
                       score,
                       bestScore,
                       bestCount,
                       newBest ? "  BEST" :
                       better ? "  best" :
                       same ? "  same" : "");
            }
        }

        solution.resize(mergeSolutionSize);

        return keep;
    };

    check(rects, true);

    bool destroyRequested = false;
    bool destroyed = false;
    Graphics gr;

    bool inImage = false;
    uint64_t imageX = 0;
    uint64_t imageY = 0;

    bool creating = false;
    uint64_t createX = 0;
    uint64_t createY = 0;

    bool tweaking = false;

    auto tweak = [&](uint64_t delta)
    {
        printf("Tweaking (delta = %" PRIu64 ")...\n", delta);

        beginTransaction();

        bool didSomething = false;

        for (uint64_t rectIdx = rects.size(); rectIdx > 0; )
        {
            rectIdx--;

            if (rects[rectIdx].lowX >= delta)
            {
                auto curRects = rects;
                curRects[rectIdx].lowX -= delta;
                didSomething = check(curRects, false, true) || didSomething;
            }

            if (rects[rectIdx].highX - rects[rectIdx].lowX > delta)
            {
                auto curRects = rects;
                curRects[rectIdx].lowX += delta;
                didSomething = check(curRects, false, false) || didSomething;
            }

            if (width - rects[rectIdx].highX >= delta)
            {
                auto curRects = rects;
                curRects[rectIdx].highX += delta;
                didSomething = check(curRects, false, true) || didSomething;
            }

            if (rects[rectIdx].highX - rects[rectIdx].lowX > delta)
            {
                auto curRects = rects;
                curRects[rectIdx].highX -= delta;
                didSomething = check(curRects, false, false) || didSomething;
            }

            if (rects[rectIdx].lowY >= delta)
            {
                auto curRects = rects;
                curRects[rectIdx].lowY -= delta;
                didSomething = check(curRects, false, true) || didSomething;
            }

            if (rects[rectIdx].highY - rects[rectIdx].lowY > delta)
            {
                auto curRects = rects;
                curRects[rectIdx].lowY += delta;
                didSomething = check(curRects, false, false) || didSomething;
            }

            if (height - rects[rectIdx].highY >= delta)
            {
                auto curRects = rects;
                curRects[rectIdx].highY += delta;
                didSomething = check(curRects, false, true) || didSomething;
            }

            if (rects[rectIdx].highY - rects[rectIdx].lowY > delta)
            {
                auto curRects = rects;
                curRects[rectIdx].highY -= delta;
                didSomething = check(curRects, false, false) || didSomething;
            }
        }

        endTransaction();

        printf("done.\n");

        if (didSomething)
        {
            gr.requestPaint();
        }
        else
        {
            if (tweaking)
            {
                tweaking = false;
                printf("tweaking = false\n");
            }
        }
    };

    auto tryDelete = [&]()
    {
        printf("Deleting...\n");

        beginTransaction();

        bool didSomething = false;

        for (uint64_t rectIdx = rects.size(); rectIdx > 0; )
        {
            rectIdx--;

            auto curRects = rects;
            curRects.erase(curRects.begin() + rectIdx);
            didSomething = check(curRects, false, true) || didSomething;
        }

        endTransaction();

        printf("done.\n");

        if (didSomething)
        {
            gr.requestPaint();
        }
    };

    auto drawImage = [&](int32_t x, int32_t y, const Image& image)
    {
        size_t imageWidth = image.getWidth();
        size_t imageHeight = image.getHeight();
        vector<uint32_t> data(imageWidth * imageHeight);
        for (size_t y = 0; y < imageHeight; y++)
        {
            for (size_t x = 0; x < imageWidth; x++)
            {
                Color color = image(x, y);
                uint32_t val =
                    ((uint32_t)color.r << 16) |
                    ((uint32_t)color.g << 8) |
                    ((uint32_t)color.b << 0);
                data[(imageHeight - 1 - y) * imageWidth + x] = val;
            }
        }

        gr.drawBitmap(x, y, data.data(), imageWidth, imageHeight);
    };

    auto onDestroy = [&]()
    {
        destroyed = true;
    };
    gr.setOnDestroy(onDestroy);

    auto onPaint = [&]()
    {
        gr.beginPaint(true);

        uint32_t grWidth = gr.getWidth();
        uint32_t grHeight = gr.getHeight();
        gr.fillRect(0, 0, grWidth, grHeight, 0x404040);

        string text = strprintf("Problem %2" PRIu64 ": %6" PRIu64 " = %6" PRIu64 " + %6" PRIu64 "", id, score, similarity, cost);

        gr.setFont(FontSpacing::Monospace, FontFace::SansSerif,
                   18, false, false);

        gr.drawText(10, 10, text.c_str(), 0xffffff, HAlign::Left, VAlign::Top);

        //drawImage(50, 50, problem.initialImage);
        drawImage(50, 50, errorImage);
        drawImage(500, 50, problem.targetImage);
        drawImage(950, 50, result);

        //gr.setLineWidth(3);
        for (Rect rect : rects)
        {
            auto doLine = [&](uint64_t x1,
                              uint64_t y1,
                              uint64_t x2,
                              uint64_t y2)
            {
                int32_t gx1 = 500 + x1;
                int32_t gy1 = 450 - 1 - y1;
                int32_t gx2 = 500 + x2;
                int32_t gy2 = 450 - 1 - y2;
                gr.setLineStyle(LineStyle::Solid);
                gr.drawLine(gx1, gy1, gx2, gy2, 0xffffff);
                gr.setLineStyle(LineStyle::Dash);
                gr.drawLine(gx1, gy1, gx2, gy2, 0x000000);
            };

            doLine(rect.lowX, rect.lowY, rect.highX, rect.lowY);
            doLine(rect.lowX, rect.highY, rect.highX, rect.highY);
            doLine(rect.lowX, rect.lowY, rect.lowX, rect.highY);
            doLine(rect.highX, rect.lowY, rect.highX, rect.highY);
        }

        gr.endPaint();
    };
    gr.setOnPaint(onPaint);

    auto onKey = [&](Key key)
    {
        switch (key)
        {
        case Key::Q:
        case Key::ShiftQ:
        {
            if (!destroyRequested)
            {
                destroyRequested = true;
                gr.destroy();
            }
            break;
        }
        case Key::Space:
        {
            tweaking = true;
            printf("tweaking = true\n");
            break;
        }
        case Key::Esc:
        {
            tweaking = false;
            printf("tweaking = false\n");
            break;
        }
        case Key::C:
        case Key::ShiftC:
        {
            if (!rects.empty())
            {
                beginTransaction();
                rects.clear();
                check(rects, true);
                endTransaction();
                gr.requestPaint();
            }
            break;
        }
        case Key::D:
        case Key::ShiftD:
        {
            if (!rects.empty())
            {
                tryDelete();
            }
            break;
        }
        case Key::Backspace:
        {
            if (!rects.empty())
            {
                beginTransaction();
                rects.pop_back();
                check(rects, true);
                endTransaction();
                gr.requestPaint();
            }
            break;
        }
        case Key::L:
        case Key::ShiftL:
        {
            beginTransaction();
            rects.clear();
            {
                Canvas tmpCanvas;
                if (!eval(problem, best.getSolution(), tmpCanvas, nullptr, nullptr, nullptr, &rects, &msg))
                {
                    printf("%s\n", msg.c_str());
                    rects.clear();
                }
            }
            check(rects, true);
            endTransaction();
            gr.requestPaint();
            break;
        }
        case Key::Key1: tweak(1); break;
        case Key::Key2: tweak(2); break;
        case Key::Key3: tweak(4); break;
        case Key::Key4: tweak(8); break;
        case Key::Key5: tweak(16); break;
        case Key::Key6: tweak(32); break;
        case Key::Key7: tweak(64); break;
        case Key::Key8: tweak(128); break;
        case Key::Key9: tweak(256); break;
        case Key::CtrlLeft:
        {
            if (!undoStack.empty())
            {
                redoStack.push_back(rects);
                rects = undoStack.back();
                undoStack.pop_back();
                check(rects, true);
                gr.requestPaint();
            }
            break;
        }
        case Key::CtrlRight:
        {
            if (!redoStack.empty())
            {
                undoStack.push_back(rects);
                rects = redoStack.back();
                redoStack.pop_back();
                check(rects, true);
                gr.requestPaint();
            }
            break;
        }
        default: ;
        }
    };
    gr.setOnKey(onKey);

    auto onMouse = [&](MouseEvent event, int32_t mouseX, int32_t mouseY)
    {
        if (mouseX >= 500 && mouseX < 900 &&
            mouseY >= 50 && mouseY < 450)
        {
            inImage = true;
            imageX = mouseX - 500;
            imageY = 450 - 1 - mouseY;
        }
        else
        {
            inImage = false;
        }

        switch (event)
        {
        case MouseEvent::RButtonDown:
        {
            if (inImage)
            {
                creating = true;
                createX = imageX;
                createY = imageY;
            }
            break;
        }
        case MouseEvent::RButtonUp:
        {
            if (creating)
            {
                if (inImage)
                {
                    beginTransaction();
                    Rect rect;
                    rect.lowX = std::min(createX, imageX);
                    rect.highX = std::max(createX, imageX) + 1;
                    rect.lowY = std::min(createY, imageY);
                    rect.highY = std::max(createY, imageY) + 1;
                    rects.push_back(rect);
                    check(rects, true);
                    endTransaction();
                    gr.requestPaint();
                }
                creating = false;
            }
            break;
        }
        default: ;
        }
    };
    gr.setOnMouse(onMouse);

    if (!gr.create(1400, 500, "Edit", &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        return 1;
    }

    while (true)
    {
        while (Graphics::update(!tweaking))
        {
            if (destroyed)
            {
                break;
            }
        }

        if (destroyed)
        {
            break;
        }

        if (tweaking)
        {
            tweak(1);
        }
        else
        {
            sleepMS(10);
        }
    }

    return 0;
}
