#include "Common.hpp"
#include "ParseUtils.hpp"
#include "Problem.hpp"
#include "Best.hpp"
#include "CalcScore.hpp"
#include "FindColor.hpp"
#include "Merge.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: rect [<options>] <id> [<delta>]\n");
    printf("  <id>  Problem id\n");
    printf("  <delta>\n");
    printf("        Rectangle coordinate granularity (default: 16)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;
    bool gotDelta = false;

    bool help = false;
    bool saveChanges = true;
    uint64_t id = 0;
    uint64_t delta = 16;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveChanges = false;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > 1000)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else if (!gotDelta)
        {
            if (!parseU64(strArg, &delta) || delta < 1 || delta > 1000)
            {
                usage();
                return 1;
            }
            gotDelta = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    string msg;

    Problem problem;
    if (!loadProblem(id, &problem))
    {
        printf("Error loading problem\n");
        return 1;
    }

    Best best;
    best.load(id);
    uint64_t bestScore = (uint64_t)-1;
    uint64_t bestCount = 0;

    Canvas canvas(problem.initialCanvas);
    Solution solution;

    if (!merge(problem, canvas, solution))
    {
        printf("Merge failed\n");
        return 1;
    }

    size_t mergeSolutionSize = solution.size();

    uint64_t mergeBlockNum = (uint64_t)-1;
    for (uint64_t curBlockNum = 0; curBlockNum < canvas.blocks.size(); curBlockNum++)
    {
        auto& block = canvas.blocks[curBlockNum];
        if (!block.valid) continue;
        if (mergeBlockNum != (uint64_t)-1)
        {
            printf("Not a single block");
            return 1;
        }
        mergeBlockNum = curBlockNum;
    }

    uint64_t width = problem.width;
    uint64_t height = problem.height;

    vector<Coord> coordsIn;
    vector<Coord> coordsOut;
    coordsIn.reserve(width * height);
    coordsOut.reserve(width * height);

    for (uint64_t lowY = delta; lowY < height; lowY += delta)
    {
        for (uint64_t highY = lowY + delta; highY < height; highY += delta)
        {
            for (uint64_t lowX = delta; lowX < width; lowX += delta)
            {
                for (uint64_t highX = lowX + delta; highX < width; highX += delta)
                {
                    coordsIn.clear();
                    coordsOut.clear();
                    for (uint64_t y = 0; y < height; y++)
                    {
                        for (uint64_t x = 0; x < width; x++)
                        {
                            if (y >= lowY && y < highY &&
                                x >= lowX && x < highX)
                            {
                                coordsIn.emplace_back(x, y);
                            }
                            else
                            {
                                coordsOut.emplace_back(x, y);
                            }
                        }
                    }

                    if (coordsOut.empty()) continue;

                    Color colorIn = findColor(problem, coordsIn);
                    Color colorOut = findColor(problem, coordsOut);

                    if (colorOut == colorIn) continue;

                    for (uint64_t pass = 0; pass < 2; pass++)
                    {
                        solution.resize(mergeSolutionSize);

                        if (pass == 1)
                        {
                            auto& move = solution.emplace_back();
                            move.moveKind = MoveKind::Color;
                            move.blockNums[0] = mergeBlockNum;
                            move.color = colorOut;
                        }

                        {
                            auto& move = solution.emplace_back();
                            move.moveKind = MoveKind::CutXY;
                            move.blockNums[0] = mergeBlockNum;
                            move.x = lowX;
                            move.y = lowY;
                        }

                        {
                            auto& move = solution.emplace_back();
                            move.moveKind = MoveKind::CutXY;
                            move.blockNums[0] = mergeBlockNum + 1 + 2;
                            move.x = highX;
                            move.y = highY;
                        }

                        {
                            auto& move = solution.emplace_back();
                            move.moveKind = MoveKind::Color;
                            move.blockNums[0] = mergeBlockNum + 1 + 4 + 0;
                            move.color = colorIn;
                        }

                        uint64_t score;
                        if (!calcScore(problem, solution, nullptr, &score, nullptr, nullptr, nullptr, &msg))
                        {
                            printf("%s\n", msg.c_str());
                            continue;
                        }

                        bool better = score < bestScore;
                        bool same = score == bestScore;
                        bool newBest = better && saveChanges && best.update(solution);

                        if (better)
                        {
                            bestScore = score;
                            bestCount = 1;
                        }
                        else if (same)
                        {
                            bestCount++;
                        }

                        if (better || same)
                        {
                            printf("%2" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "%s\n",
                                   id,
                                   score,
                                   bestScore,
                                   bestCount,
                                   newBest ? "  BEST" :
                                   better ? "  best" :
                                   same ? "  same" : "");
                        }
                    }
                }
            }
        }
    }

    return 0;
}
