#include "Common.hpp"
#include "ParseUtils.hpp"
#include "Problem.hpp"
#include "Best.hpp"
#include "CalcScore.hpp"
#include "Merge.hpp"
#include "Rects.hpp"
#include "FindColors.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: rects [<options>] <id> [<delta>]\n");
    printf("  <id>  Problem id\n");
    printf("  <delta>\n");
    printf("        Rectangle coordinate granularity (default: 16)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;
    bool gotDelta = false;

    bool help = false;
    bool saveChanges = true;
    uint64_t id = 0;
    uint64_t delta = 16;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveChanges = false;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > 1000)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else if (!gotDelta)
        {
            if (!parseU64(strArg, &delta) || delta < 1 || delta > 1000)
            {
                usage();
                return 1;
            }
            gotDelta = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    string msg;

    Problem problem;
    if (!loadProblem(id, &problem))
    {
        printf("Error loading problem\n");
        return 1;
    }

    Best best;
    best.load(id);
    uint64_t bestScore = (uint64_t)-1;
    uint64_t bestCount = 0;

    Canvas canvas(problem.initialCanvas);
    Solution solution;

    if (!merge(problem, canvas, solution))
    {
        printf("Merge failed\n");
        return 1;
    }

    size_t mergeSolutionSize = solution.size();

    uint64_t width = problem.width;
    uint64_t height = problem.height;

    vector<Rect> rects;

    bool improved = true;
    while (improved)
    {
        printf("=== %" PRIuZ " rects ===\n", rects.size() + 1);

        improved = false;

        Rect bestRect;

        for (uint64_t lowY = 0; lowY < height; lowY += delta)
        {
            for (uint64_t highY = lowY + delta; highY <= height; highY += delta)
            {
                for (uint64_t lowX = 0; lowX < width; lowX += delta)
                {
                    for (uint64_t highX = lowX + delta; highX <= width; highX += delta)
                    {
                        Rect rect(lowX, lowY, highX, highY);
                        rects.push_back(rect);

                        {
                            auto& move = solution.emplace_back();
                            move.moveKind = MoveKind::Color;
                            move.blockNums[0] = canvas.blocks.size() - 1;
                            //move.color = ...;
                        }

                        Canvas curCanvas(canvas);
                        makeRects(problem, rects, curCanvas, solution);

                        findColors(problem, solution);

                        if (problem.initialCanvas.blocks.size() == 1 &&
                            solution[mergeSolutionSize].color == problem.initialImage(0, 0))
                        {
                            solution.erase(solution.begin() + mergeSolutionSize);
                        }

                        uint64_t score;
                        if (!calcScore(problem, solution, nullptr, &score, nullptr, nullptr, nullptr, &msg))
                        {
                            fprintf(stderr, "%s\n", msg.c_str());
                            exit(1);
                        }

                        bool better = score < bestScore;
                        bool same = score == bestScore;
                        bool newBest = better && saveChanges && best.update(solution);

                        if (better)
                        {
                            bestScore = score;
                            bestCount = 1;
                            bestRect = rect;
                            improved = true;
                        }
                        else if (same)
                        {
                            bestCount++;
                        }

                        if (better || same)
                        {
                            printf("%2" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "%s\n",
                                   id,
                                   score,
                                   bestScore,
                                   bestCount,
                                   newBest ? "  BEST" :
                                   better ? "  best" :
                                   same ? "  same" : "");
                        }

                        rects.pop_back();
                        solution.resize(mergeSolutionSize);
                    }
                }
            }
        }

        if (improved)
        {
            rects.push_back(bestRect);
        }
    }

    return 0;
}
