#include "Common.hpp"
#include "ParseUtils.hpp"
#include "Problem.hpp"
#include "Best.hpp"
#include "CalcScore.hpp"
#include "Merge.hpp"
#include "Rects.hpp"
#include "FindColors.hpp"
#include "Xoshiro.hpp"
#include "FastRandom.hpp"
#include "TimeUtils.hpp"
#include "ProcessId.hpp"

using std::string;
using std::vector;

typedef Xoshiro256StarStar Gen;

void usage()
{
    printf("Usage: rects [<options>] <id>\n");
    printf("  <id>  Problem id\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    bool saveChanges = true;
    uint64_t id = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveChanges = false;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > 1000)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    string msg;

    uint32_t seed1 = 1;
    uint32_t seed2 = 1;

    {
        uint64_t timeMS = getTimeMS();
        uint32_t pid = getProcessId();

        seed1 = (uint32_t)timeMS;
        seed2 = (uint32_t)pid ^ (uint32_t)(timeMS >> 32);
    }

    std::seed_seq seq{seed1, seed2};
    Gen gen(seq);

    Problem problem;
    if (!loadProblem(id, &problem))
    {
        printf("Error loading problem\n");
        return 1;
    }

    Best best;
    best.load(id);
    uint64_t bestScore = (uint64_t)-1;
    uint64_t bestCount = 0;

    Canvas canvas(problem.initialCanvas);
    Solution solution;

    if (!merge(problem, canvas, solution))
    {
        printf("Merge failed\n");
        return 1;
    }

    size_t mergeSolutionSize = solution.size();

    uint64_t width = problem.width;
    uint64_t height = problem.height;

    bool improved = true;
    vector<Rect> bestRects;

    auto check = [&](const vector<Rect>& rects) -> uint64_t
    {
        {
            auto& move = solution.emplace_back();
            move.moveKind = MoveKind::Color;
            move.blockNums[0] = canvas.blocks.size() - 1;
            //move.color = ...;
        }

        Canvas curCanvas(canvas);
        makeRects(problem, rects, curCanvas, solution);

        findColors(problem, solution);

        if (problem.initialCanvas.blocks.size() == 1 &&
            !problem.hasSourceImage &&
            solution[mergeSolutionSize].color == problem.initialImage(0, 0))
        {
            solution.erase(solution.begin() + mergeSolutionSize);
        }

        uint64_t score;
        if (!calcScore(problem, solution, nullptr, &score, nullptr, nullptr, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        bool better = score < bestScore;
        bool same = score == bestScore;
        bool newBest = better && saveChanges && best.update(solution);

        if (better)
        {
            bestScore = score;
            bestCount = 1;
            bestRects = rects;
            improved = true;
        }
        else if (same)
        {
            bestCount++;
        }

        if (better || same)
        {
            printf("%2" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "%s\n",
                   id,
                   score,
                   bestScore,
                   bestCount,
                   newBest ? "  BEST" :
                   better ? "  best" :
                   same ? "  same" : "");
        }

        solution.resize(mergeSolutionSize);

        return score;
    };

    while (improved)
    {
        printf("=== %" PRIuZ " rects ===\n", bestRects.size() + 1);

        improved = false;

        vector<Rect> startRects(bestRects);

        for (uint64_t i = 0; i < (improved ? 10 : 20); i++)
        {
            printf("i = %" PRIu64 "\n", i);

            uint64_t lowY;
            uint64_t highY;
            while (true)
            {
                lowY = fastRandom(height, gen);
                highY = fastRandom(height, gen) + 1;
                if (lowY < highY) break;
            }
            uint64_t lowX;
            uint64_t highX;
            while (true)
            {
                lowX = fastRandom(height, gen);
                highX = fastRandom(height, gen) + 1;
                if (lowX < highX) break;
            }

            vector<Rect> rects(startRects);
            rects.emplace_back(lowX, lowY, highX, highY);

            uint64_t score = check(rects);

            for (uint64_t delta = 64; delta != 0; delta /= 2)
            {
                while (true)
                {
                    bool didSomething = false;

                    for (uint64_t rectIdx = rects.size(); rectIdx > 0; )
                    {
                        rectIdx--;

                        vector<Rect> curRects(rects);
                        auto& curRect = curRects[rectIdx];

                        if (curRect.lowX >= delta)
                        {
                            curRect.lowX -= delta;
                            uint64_t curScore = check(curRects);
                            if (curScore < score)
                            {
                                rects = curRects;
                                score = curScore;
                                didSomething = true;
                            }
                            else
                            {
                                curRect.lowX += delta;
                            }
                        }

                        if (curRect.highX - curRect.lowX > delta)
                        {
                            curRect.lowX += delta;
                            uint64_t curScore = check(curRects);
                            if (curScore < score)
                            {
                                rects = curRects;
                                score = curScore;
                                didSomething = true;
                            }
                            else
                            {
                                curRect.lowX -= delta;
                            }
                        }

                        if (width - curRect.highX >= delta)
                        {
                            curRect.highX += delta;
                            uint64_t curScore = check(curRects);
                            if (curScore < score)
                            {
                                rects = curRects;
                                score = curScore;
                                didSomething = true;
                            }
                            else
                            {
                                curRect.highX -= delta;
                            }
                        }

                        if (curRect.highX - curRect.lowX > delta)
                        {
                            curRect.highX -= delta;
                            uint64_t curScore = check(curRects);
                            if (curScore < score)
                            {
                                rects = curRects;
                                score = curScore;
                                didSomething = true;
                            }
                            else
                            {
                                curRect.highX += delta;
                            }
                        }

                        if (curRect.lowY >= delta)
                        {
                            curRect.lowY -= delta;
                            uint64_t curScore = check(curRects);
                            if (curScore < score)
                            {
                                rects = curRects;
                                score = curScore;
                                didSomething = true;
                            }
                            else
                            {
                                curRect.lowY += delta;
                            }
                        }

                        if (curRect.highY - curRect.lowY > delta)
                        {
                            curRect.lowY += delta;
                            uint64_t curScore = check(curRects);
                            if (curScore < score)
                            {
                                rects = curRects;
                                score = curScore;
                                didSomething = true;
                            }
                            else
                            {
                                curRect.lowY -= delta;
                            }
                        }

                        if (height - curRect.highY >= delta)
                        {
                            curRect.highY += delta;
                            uint64_t curScore = check(curRects);
                            if (curScore < score)
                            {
                                rects = curRects;
                                score = curScore;
                                didSomething = true;
                            }
                            else
                            {
                                curRect.highY -= delta;
                            }
                        }

                        if (curRect.highY - curRect.lowY > delta)
                        {
                            curRect.highY -= delta;
                            uint64_t curScore = check(curRects);
                            if (curScore < score)
                            {
                                rects = curRects;
                                score = curScore;
                                didSomething = true;
                            }
                            else
                            {
                                curRect.highY += delta;
                            }
                        }
                    }

                    if (!didSomething)
                    {
                        break;
                    }
                }
            }
        }
    }

    return 0;
}
