#include "Common.hpp"
#include "Graphics.hpp"
#include "ProblemCount.hpp"
#include "Problem.hpp"
#include "Best.hpp"
#include "StringUtils.hpp"

using std::string;
using std::vector;

int main(int argc, char *argv[])
{
    uint64_t id = 1;
    Problem problem;

    string msg;
    if (!loadProblem(id, &problem, &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        return 1;
    }

    Best best;
    auto update = [&]()
    {
        best.load(id);
    };

    update();

    bool destroyRequested = false;
    bool destroyed = false;
    Graphics gr;

    auto drawImage = [&](int32_t x, int32_t y, const Image& image)
    {
        size_t imageWidth = image.getWidth();
        size_t imageHeight = image.getHeight();
        vector<uint32_t> data(imageWidth * imageHeight);
        for (size_t y = 0; y < imageHeight; y++)
        {
            for (size_t x = 0; x < imageWidth; x++)
            {
                Color color = image(x, y);
                uint32_t val =
                    ((uint32_t)color.r << 16) |
                    ((uint32_t)color.g << 8) |
                    ((uint32_t)color.b << 0);
                data[(imageHeight - 1 - y) * imageWidth + x] = val;
            }
        }

        gr.drawBitmap(x, y, data.data(), imageWidth, imageHeight);
    };

    auto onDestroy = [&]()
    {
        destroyed = true;
    };
    gr.setOnDestroy(onDestroy);

    auto onPaint = [&]()
    {
        gr.beginPaint(true);

        uint32_t width = gr.getWidth();
        uint32_t height = gr.getHeight();
        gr.fillRect(0, 0, width, height, 0x404040);

        string text = strprintf("Problem %2" PRIu64 ": %6" PRIu64 " = %6" PRIu64 " + %6" PRIu64 "", id, best.getScore(), best.getSimilarity(), best.getCost());

        gr.setFont(FontSpacing::Monospace, FontFace::SansSerif,
                   18, false, false);

        gr.drawText(10, 10, text.c_str(), 0xffffff, HAlign::Left, VAlign::Top);

        drawImage(50, 50, problem.initialImage);
        drawImage(50 + problem.initialImage.getWidth() + 50, 50, problem.targetImage);
        drawImage(50 + problem.initialImage.getWidth() + 50 + problem.targetImage.getWidth() + 50, 50, best.getResult());

        gr.endPaint();
    };
    gr.setOnPaint(onPaint);

    auto onKey = [&](Key key)
    {
        string msg;

        switch (key)
        {
        case Key::Esc:
        case Key::Q:
        case Key::ShiftQ:
        {
            if (!destroyRequested)
            {
                destroyRequested = true;
                gr.destroy();
            }
            break;
        }
        case Key::Right:
        {
            if (id < problemCount)
            {
                id++;
                if (!loadProblem(id, &problem, &msg))
                {
                    fprintf(stderr, "%s\n", msg.c_str());
                    exit(1);
                }
                update();
                gr.requestPaint();
            }
            break;
        }
        case Key::Left:
        {
            if (id > 1)
            {
                id--;
                if (!loadProblem(id, &problem, &msg))
                {
                    fprintf(stderr, "%s\n", msg.c_str());
                    exit(1);
                }
                update();
                gr.requestPaint();
            }
            break;
        }
        case Key::CtrlRight:
        {
            if (id < problemCount)
            {
                id += std::min((uint64_t)10, problemCount - id);
                if (!loadProblem(id, &problem, &msg))
                {
                    fprintf(stderr, "%s\n", msg.c_str());
                    exit(1);
                }
                update();
                gr.requestPaint();
            }
            break;
        }
        case Key::CtrlLeft:
        {
            if (id > 1)
            {
                id -= std::min((uint64_t)10, id - 1);
                if (!loadProblem(id, &problem, &msg))
                {
                    fprintf(stderr, "%s\n", msg.c_str());
                    exit(1);
                }
                update();
                gr.requestPaint();
            }
            break;
        }
        default: ;
        }
    };
    gr.setOnKey(onKey);

    if (!gr.create(1400, 500, "Show", &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        return 1;
    }

    while (!destroyed)
    {
        if (Graphics::update(true))
        {
            continue;
        }
    }

    return 0;
}
