#include "Common.hpp"
#include "ProblemCount.hpp"
#include "Problem.hpp"
#include "Best.hpp"
#include "CalcScore.hpp"
#include "FindColor.hpp"
#include "Merge.hpp"

using std::string;
using std::vector;

void solveSolid(const Problem& problem,
                Canvas& canvas,
                Solution& solution)
{
    uint64_t width = problem.width;
    uint64_t height = problem.height;

    vector<Coord> coords;
    coords.reserve(width * height);

    for (uint64_t y = 0; y < height; y++)
    {
        for (uint64_t x = 0; x < width; x++)
        {
            coords.emplace_back(x, y);
        }
    }

    Color color = findColor(problem, coords);

    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::Color;
        move.blockNums[0] = canvas.blocks.size() - 1;
        move.color = color;
    }
}

void usage()
{
    printf("Usage: solid [<options>]\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
}

int main(int argc, char *argv[])
{
    bool help = false;
    bool saveChanges = true;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveChanges = false;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    string msg;

    for (uint64_t id = 1; id <= problemCount; id++)
    {
        Problem problem;
        if (!loadProblem(id, &problem, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Best best;
        best.load(id);
        uint64_t bestScore = best.getScore();

        Canvas canvas(problem.initialCanvas);
        Solution solution;

        if (!merge(problem, canvas, solution))
        {
            printf("Merge failed\n");
            continue;
        }

        solveSolid(problem, canvas, solution);

        uint64_t score;
        if (!calcScore(problem, solution, nullptr, &score, nullptr, nullptr, nullptr, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        bool better = score < bestScore;
        bool same = score == bestScore;
        bool newBest = better && saveChanges && best.update(solution);

        printf("%2" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "%s\n",
               id,
               bestScore,
               score,
               newBest ? "  BEST" :
               better ? "  best" :
               same ? "  same" : "");
    }

    return 0;
}
